/*  FrameBuffer Driver for SiS 530 (6306) Graphics chips
 *
 *
 *
 *
 * (c) Christopher Wright 2002 <cwright@softpixel.com>
 */
 
 /* devlog:
  *  2002.10.22 - initial complete rewrite
  *
  */
  
  /*  This driver is based partly on sisfb, vesafb, 
   *  matroxfb, svgalib 1.4.3, xfree 4.1, and others 
   */
  
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/selection.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/fb.h>
#include <linux/tty.h>
#include <linux/console.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/slab.h>

#include <asm/io.h>
#include <asm/mtrr.h>

#include <video/fbcon.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>
#include <video/fbcon-cfb32.h>

#define VENDOR	0x1039	// SiS
#define DEVICE	0x6306	// 530 chip (6306)

#define DEBUG(x...)	printk(KERN_DEBUG "SiS530: " x)

#define sis530wb(reg,data)	(outb(data,card->vgaBase+reg))
#define sis530ww(reg,data)	(outw(data,card->vgaBase+reg))
#define sis530wl(reg,data)	(outl(data,card->vgaBase+reg))
#define sis530rb(reg)		(inb(card->vgaBase+reg))
#define sis530rw(reg)		(inw(card->vgaBase+reg))

#define SEQ_W	0x14
#define SEQ_R	0x15
#define CRT_W	0x24
#define CRT_R	0x25
#define GCR_W	0x1E
#define GCR_R	0x1F
#define ATR_W	0x11
#define ATR_R	0x11

MODULE_AUTHOR("Christopher Wright, cwright@softpixel.com");
MODULE_DESCRIPTION("SiS530 Framebuffer Driver");
MODULE_LICENSE("GPL");

static int	currentConsole=0;
static struct	display	display;
static struct	fb_info fb_info;

typedef struct SiS530_Card
{
	u32		videoBase;
	u32		mmioBase;
	u32		vgaBase;
	u8		*videoPtr;
	u8		*mmioPtr;
	u32		memSize;
	struct pci_dev	*pdev;
	u32		use;
}SiS530Card;


SiS530Card *card=NULL;	//mono-head for now..

/* Frame buffer device API */
static int sis530_open(struct fb_info *info,int user)
{
	DEBUG("Open!\n");
	card->use++;
	return 0;
}

static int sis530_release(struct fb_info *info, int user)
{
	DEBUG("Release!\n");
	card->use--;
	return 0;
}

static int sis530_get_fix(struct fb_fix_screeninfo *fix, int con,struct fb_info *info)
{
	DEBUG("Get Fix\n");
	memset(fix,0,sizeof(struct fb_fix_screeninfo));

	strcpy(fix->id,"SiS530 (6306)");
	fix->smem_start=(unsigned long)card->videoPtr;
	fix->smem_len=card->memSize;
	fix->type=FB_TYPE_PACKED_PIXELS;	//FIXME (curcon's actual type)
	fix->visual=FB_VISUAL_TRUECOLOR;	//FIXME (curcon's actual visual)
	fix->xpanstep=0;	
	fix->ypanstep=0;	
	fix->ywrapstep=0;
	fix->line_length=1280;	//FIXME (curcon's width * depth)
	fix->mmio_start=(unsigned long)card->mmioPtr;
	fix->mmio_len=pci_resource_len(card->pdev,1);
	fix->accel=FB_ACCEL_NONE;

	return 0;
}

static int sis530_get_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
	DEBUG("Get Var\n");
	
	return 0;
}

static int sis530_set_var(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
	DEBUG("Set Var\n");
	
	return -EINVAL;
}

static int sis530_pan_display(struct fb_var_screeninfo *var, int con, struct fb_info *info)
{
	DEBUG("Pan Display\n");
	
	return -EINVAL;
}

static int sis530_get_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
	DEBUG("Get Cmap\n");
	
	return 0;
}

static int sis530_set_cmap(struct fb_cmap *cmap, int kspc, int con, struct fb_info *info)
{
	DEBUG("Set Cmap\n");
	
	return 0;
}

/* console interface */
static int sis530_con_switch(int con, struct fb_info *fb)
{
	DEBUG("switch to %d\n",con);
	return 0;
}
		
static int sis530_con_updatevar(int con, struct fb_info *fb)
{
	DEBUG("updatevar to %d\n",con);
	return 0;
}
		
static void sis530_con_blank(int blank, struct fb_info *fb)
{
	DEBUG("Blank to %d\n",blank);
}

static struct fb_ops sis530_ops = {
	owner:		THIS_MODULE,
	fb_open:	sis530_open,
	fb_release:	sis530_release,
	fb_get_fix:	sis530_get_fix,
	fb_get_var:	sis530_get_var,
	fb_set_var:	sis530_set_var,
	fb_get_cmap:	sis530_get_cmap,
	fb_set_cmap:	sis530_set_cmap,
	fb_pan_display:	sis530_pan_display,
};


/* _find finds the sis530 device(s), and initializes the card struct */
/* it also unlocks the device (0x86), and completely messes up with multi-head */
/* I dont know if one can multi-head a sis530, so it may not be a problem.. */
int __init sis530_find(void)
{
	struct pci_dev *pdev=NULL;
	int i;
	int devs=0;
	
	DEBUG("Finding SiS530...\n");
	
	while((pdev=pci_find_device(VENDOR,DEVICE,pdev)))
	{
		DEBUG("Found!\n");
		if(pci_enable_device(pdev)==0)
		{
			DEBUG("Yay, enabled SiS530!\n");
			
			if((card=(SiS530Card*)kmalloc(sizeof(SiS530Card),GFP_KERNEL))==NULL)
			{
				DEBUG("Out of Memory\n");
				return -ENOMEM;
			}
			DEBUG("Setting Values...\n");
			card->pdev=pdev;
			card->use=0;	//not used (just made :)
			
			card->videoBase = pci_resource_start(pdev,0);
			DEBUG("   VideoBase:  %x\n",card->videoBase);
			card->mmioBase	= pci_resource_start(pdev,1);
			DEBUG("   mmioBase:   %x\n",card->mmioBase);
			card->vgaBase	= pci_resource_start(pdev,2)+0x30;
			DEBUG("   vgaBase:    %x\n",card->vgaBase);
			DEBUG("All set up!\n");
			
			/* here, we should request_mem_region() and stuff, but we don't */
			
			sis530wb(SEQ_W,5);
			sis530wb(SEQ_W,0x86);	// unlock extended registers
			
			/* calculate memory size */
			sis530wb(SEQ_W,0xc);
			i=sis530rb(SEQ_R)>>1;
			i=(i&3) | ( (i&8) >> 1 );
			if( ( ((i&4) && (i&3)==0) )|| ( !(i&4)&&(i&3)==3 ) )
			{
				DEBUG("Invalid Memory Size\n");
				kfree(card);
				return -ENODEV;
			}
			DEBUG("Memory: %d bytes\n",(1024*1024)<<i);
			card->memSize=(1024*1024)<<i;
		}
		else
		{
			DEBUG("...But enabling it failed\n");
			return -EBUSY;
		}
		++devs;
		if(devs>1)
		{
			DEBUG("Uhoh, multi-head.  I'm not prepared to deal with this yet...\n");
			DEBUG("Expect some bugs and some crazy behaviour (only 1 card working)\n");
		}
			
	}
	
	return 0;	//success
}

/* master initalization function */
int __init sis530_init(void)
{
	int ret;
	DEBUG("Initializing Device...\n");

	if(! (ret=sis530_find()) )
	{
		DEBUG("Device found and initialized and everything.  good stuff\n");
	}
	else
	{
		DEBUG("Something bad happened (no device?).  sorry\n");
		return -ENODEV;
	}
	
	/* ok, device is found and what not, let's set up some fb stuff.*/
#ifdef DEVELOPING	
	strcpy(fb_info.modename,"SiS Framebuffer");
	fb_info.changevar = NULL;
	fb_info.node = -1;
	fb_info.fbops=&sis530_ops;
	fb_info.disp=&display;
	fb_info.switch_con=&sis530_con_switch;
	fb_info.updatevar=&sis530_con_updatevar;
	fb_info.blank=&sis530_con_blank;
	fb_info.flags=FBINFO_FLAG_DEFAULT;	//1 if module, else 0.
		
	DEBUG("Registering framebuffer...  wish me luck\n");
	if( register_framebuffer(&fb_info) < 0)
	{
		kfree(card);
		return -EINVAL;
	}
#endif
	
	return ret;
}

#ifdef MODULE
static int __init sis530_module_init(void)
{	
	DEBUG("Initializing Module...\n");
	
	if(!pci_present())
	{
		DEBUG("This driver requires a PCI bus\n");
		return -ENODEV;
	}
	
	/* find pci device */
	if(!sis530_init())
	{
		DEBUG("All loaded and ready to go!\n");
		return 0;
	}
	DEBUG("Something went wrong somewhere.\n");
	return -ENODEV;
}

static void __exit sis530_module_exit(void)
{
	DEBUG("Exiting...\n");
#ifdef DEVELOPING
	unregister_framebuffer(&fb_info);
#endif
	DEBUG("unregistered framebuffer\n");
	if(card)
		kfree(card);
}

module_init(sis530_module_init);
module_exit(sis530_module_exit);
#endif	//MODULE
